----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    10:28:58 11/12/2015 
-- Design Name: 
-- Module Name:    Datapath - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Datapath is
  port(
    clk_i  : in std_logic;
    --Input
    data_i    : in std_logic_vector(7 downto 0);
    key_i     : in std_logic_vector(7 downto 0);
    --Output
    data_o    : out std_logic_vector(7 downto 0);
    --Control signals
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(1 downto 0);
    mode_i    : in std_logic;
    rc_init   : in std_logic;
    count_r   : in std_logic_vector(1 downto 0)
  );
end Datapath;

architecture Behavioral of Datapath is

  COMPONENT generate_constant_led
	PORT(
		rc_in : IN std_logic_vector(5 downto 0);
		mode_i : IN std_logic;          
		rc_out : OUT std_logic_vector(5 downto 0)
		);
	END COMPONENT;
    
  COMPONENT addconstant_led
	PORT(
		state_in : IN std_logic_vector(63 downto 0);
		rc : IN std_logic_vector(5 downto 0);          
		state_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT sbox
	PORT(
		sbox_i : IN std_logic_vector(3 downto 0);          
		sbox_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT sbox_d
	PORT(
		sbox_i : IN std_logic_vector(3 downto 0);          
		sbox_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT shiftrows
	PORT(
		state_in : IN std_logic_vector(63 downto 0);
		state_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT shiftrow_d
	PORT(
		state_in : IN std_logic_vector(63 downto 0);          
		state_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT mix_col
	PORT(
		state_i : IN std_logic_vector(63 downto 0);          
		state_o : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT mixcol_d
	PORT(
		state_i : IN std_logic_vector(63 downto 0);          
		state_o : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  --regsiters
  signal key_reg      : std_logic_vector(key_size-1 downto 0);
  signal state_reg    : std_logic_vector(state_size-1 downto 0);
  signal rc_reg       : std_logic_vector(5 downto 0);
  --signals
  signal add_rk      : std_logic_vector(63 downto 0);
  signal add_const_i : std_logic_vector(63 downto 0);
  signal add_const_o : std_logic_vector(63 downto 0);
  signal sub_o       : std_logic_vector(63 downto 0);
  signal subd_o      : std_logic_vector(63 downto 0);
  signal shift_o     : std_logic_vector(63 downto 0);
  signal shiftd_o    : std_logic_vector(63 downto 0);
  signal round_o     : std_logic_vector(63 downto 0);
  signal rc_o        : std_logic_vector(5 downto 0);
  signal mix_o       : std_logic_vector(63 downto 0);
  signal mixd_o      : std_logic_vector(63 downto 0);

begin

  --key process
  K_proc : process(clk_i, key_sel)
  begin
    if rising_edge(clk_i) then
      if key_sel = "01" then
        key_reg <= key_i & key_reg(key_size-1 downto 8);
      elsif key_sel = "10" then
        key_reg <= key_reg(63 downto 0) & key_reg(127 downto 64);
      else
        key_reg <= key_reg;
      end if;
    end if;
  end process;
  
  --data process
  D_proc : process(clk_i, data_sel)
  begin
    if rising_edge(clk_i) then
      if data_sel = "01" then
        state_reg <= data_i & state_reg(state_size-1 downto 8);
        if rc_init = '0' then
          rc_reg  <= (others => '0');
        else
          rc_reg  <= "001001";
        end if;
      elsif data_sel = "10" then
        state_reg <=  round_o;
        rc_reg    <= rc_o;
      elsif data_sel = "11" then
        state_reg <= add_rk;
      else
        state_reg <= state_reg;
      end if;
    end if;
  end process;
  
  data_o <= state_reg(3 downto 0) & state_reg(7 downto 4);
  
  --addrk
  with count_r select
    add_rk <= 
      state_reg xor key_reg(63 downto 0) when "00",
      state_reg when others;
  
  with mode_i select
    add_const_i <= 
      add_rk when '0',
      subd_o when others;
      
  Inst_generate_constant_led: generate_constant_led PORT MAP(
		rc_in => rc_reg,
		mode_i => mode_i,
		rc_out => rc_o
  );

  Inst_addconstant_led: addconstant_led PORT MAP(
		state_in => add_const_i,
		rc => rc_O,
		state_out => add_const_o
	);

  sub_inst : for i in 0 to 15 generate
    Inst_sbox: sbox PORT MAP(
      sbox_i => add_const_o((i*4+3) downto i*4),
      sbox_o => sub_o((i*4+3) downto i*4)
    );
  end generate;
  
  Inst_shiftrows: shiftrows PORT MAP(
		state_in => sub_o,
		state_out => shift_o
	);
  
  Inst_shiftrow_d: shiftrow_d PORT MAP(
		state_in => mixd_o,
		state_out => shiftd_o
	);
  
  subd_inst : for i in 0 to 15 generate
    Inst_sbox_d: sbox_d PORT MAP(
      sbox_i => shiftd_o((i*4+3) downto i*4),
      sbox_o => subd_o((i*4+3) downto i*4)
    );
  end generate;
  
  Inst_mix_col: mix_col PORT MAP(
		state_i => shift_o,
		state_o => mix_o
	);
  
  Inst_mixcol_d: mixcol_d PORT MAP(
		state_i => add_rk,
		state_o => mixd_o
	);
  
  with mode_i select
    round_o <= 
      mix_o when '0',
      add_const_o when others;
  
end Behavioral;

